/* radiod.c - main file */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <signal.h>
#include "radiod.h"

/*
$Id: radiod.c,v 2.0.1.3 1997/08/05 21:36:14 anad Exp $

Program:       radiod (Radio Daemon)

Info:          this daemon is a server for radiocards Zoltrix Audio Plus
               and RadioTrack. It wait for connections on port 7289 and
	       lets the client to operate on the radiocard. This program
	       should be SUID.

Execution:     radiod

Author:        Andrzej Gab, Opole, Polska, gab@po.opole.pl

Date:          August 1997

*/

#define BACKLOG       5                 /* length of order queue            */
#define MAXHOSTNAME  32                 /* max hostname length              */
#define LOCKF        "/tmp/radiod.pid"  /* name of the lock file with pid   */
#define WORKDIR      "/tmp"             /* working directory                */
#define UMASK        077                /* umask                            */

extern long RadioType; 
extern int used;

int muted=0;
int lastvol=1;
int s4sig=0;

static void s_sig(int sig)
{
  syslog(LOG_DAEMON+LOG_NOTICE,"got %d signal, shutting down",sig);
  closelog();
  if (s4sig>0)
    close(s4sig);
  exit(0);
}

static void s_hup(int sig)
{
  int m, RootBase, RootType;

  used=0;
  defaultstation.name[0]='\0';
  defaultstation.freq=0;
  for (m=0;m<NumStations;m++)
    {
      stations[m].freq=0;
      stations[m].name[0]='\0';
    }
  RootBase=RadioBase;
  RootType=RadioType;

  yyin=fopen("/etc/radio.conf","r");
  if (yyin==NULL) { have_error(1); }

  yylex();
  
  fclose(yyin);

  RadioBase=RootBase;
  RadioType=RootType;

  syslog(LOG_DAEMON+LOG_NOTICE,"stations from /etc/radio.conf reread");

  return;
}

void main(int argc, char *argv[]) /* standard declaration of args     */
{
  int s, t;                      /* socket descriptors                */
  int i, port, n, fd, lf ;       /* working variables                 */
  struct sockaddr_in sa, isa;    /* socket address structure          */
  struct hostent *hp;            /* search name pointer wskaznik      */
  char *myname;                  /* program name pointer              */
  struct servent *sp;            /* service search return             */
  char localhost[MAXHOSTNAME+1]; /* local hostname in ASCII           */
  char pbuf[10];                 /* pid in ASCII                      */

  myname=argv[0];

/* go background :) */

  i=fork();

/* error when forking */
  if (i<0)
    {
      perror("error when forking: ");
      exit(1);
    }

/* mother process ends its execution */
  if (i)
    { 
      exit(0);
    }

/* Everything went OK - child process works in background as a server */

/* closing descriptors inherited by server */

  for (i=getdtablesize()-1; i>=0; --i)
    (void) close(i);

/* disconnecting server from controlling terminal */

  fd=open("/dev/tty",O_RDWR);
  (void) ioctl(fd,TIOCNOTTY,0);
  (void) close(fd);

/* changing server working directory to WORKDIR */

  (void) chdir(WORKDIR);

/* setting umask to UMASK */

  (void) umask(UMASK);

/* setting our server to its process group */

  (void) setpgrp(0,getpid());

/* open standard descriptors */

  fd=open("/dev/null",O_RDWR);
  (void) dup(fd);
  (void) dup(fd);

/* starting syslog() facility to log all our errors and notices */

   openlog("radiod", LOG_CONS+LOG_PID,LOG_DAEMON);
   syslog(LOG_DAEMON+LOG_NOTICE,"trying to start radiod :)");

/* locking */

  lf=open(LOCKF, O_RDWR|O_CREAT, 0640);
  if (lf<0)
    {
      syslog(LOG_DAEMON+LOG_ERR,"error opening file %s",LOCKF);
      closelog();
      exit(1);
    }
  if (flock(lf, LOCK_EX|LOCK_NB))
    {
      syslog(LOG_DAEMON+LOG_ERR,"there is probably another radiod server");
      closelog();
      exit(0);
    }

/* saving in LOCKF file process pid */

  (void) sprintf(pbuf,"%d\n",getpid());
  (void) write(lf,pbuf,strlen(pbuf));

/* Get radiod service socket number */

  if ((sp=getservbyname("radio","tcp"))==NULL)
    {
      syslog(LOG_DAEMON+LOG_ERR,"this host doesn't support radiod service: %m");
      closelog();
      exit(1);
    }

  swab(&(sp->s_port),&port,sizeof port);
  syslog(LOG_DAEMON+LOG_NOTICE,"radiod port: %d", port);

/* Get local host info */

  gethostname(localhost,MAXHOSTNAME);
  if ((hp=gethostbyname(localhost))==NULL)
    {
      syslog(LOG_DAEMON+LOG_ERR,"can't find info about local host");
      closelog();
      exit(1);
    }

/* Put socket number and our address in socket structure */

  sa.sin_port=sp->s_port;
  bcopy((char*)hp->h_addr, (char*)&sa.sin_addr, hp->h_length);
  sa.sin_family=hp->h_addrtype;

/* Open the socket for incoming connections */

  if ((s=s4sig=socket(hp->h_addrtype, SOCK_STREAM, 0))<0)
    {
      syslog(LOG_DAEMON+LOG_ERR,"socket open error: %m");
      closelog();
      exit(1);
    }

/* Bind the socket with service protocol */

  if ((bind(s, &sa, sizeof sa))<0)
    {
      syslog(LOG_DAEMON+LOG_ERR,"bind socket error: %m");
      closelog();
      exit(1);
    }

/* Set max pending connections queue length */

  listen(s, BACKLOG);

/* read config file and init radiocard */

  n=do_init();

/* signal handling install */

  (void) signal(SIGQUIT, s_sig);
  (void) signal(SIGHUP, s_hup);

/* go into endless loop waiting for connections */

  syslog(LOG_DAEMON+LOG_NOTICE,"Radiod successfully started :))");
  while (1)
    {
      i=sizeof isa;
      /* Hanging on accept() call waiting for a client */
      if ((t=accept(s, &isa, &i))<0)
	{
	  syslog(LOG_DAEMON+LOG_ERR,"connection accept() error: %m");
	  closelog();
	  close(t);
      	  exit (1);
	}
      radio(t);         /* go into radio service */
      close(t);
    }
  closelog();
}

/* print help to the client */
void s_help(int sock)
{
  char buf[BUFSIZ+1];
  sprintf(buf,"\tHELP  - this help\n");
  write(sock, buf, strlen(buf));
  sprintf(buf,"\tSTART - radio on\tOFF   - radio off\tQUIT  - quit\n");
  write(sock, buf, strlen(buf));
  sprintf(buf,"\tUP    - volume up\tDOWN  - volume down\tMUTE  - mute on/off\n");
  write(sock, buf, strlen(buf));
  sprintf(buf,"\tPRINT     - print stations in config file\n");
  write(sock, buf, strlen(buf));
  sprintf(buf,"\tTUNE n    - tune to station number n\n");
  write(sock, buf, strlen(buf));
  sprintf(buf,"\tFREQ n    - tune to given frequency\n");
  write(sock, buf, strlen(buf));
  sprintf(buf,"\tEnd of help.\n");
  write(sock, buf, strlen(buf));
  return;
}

/* switch radiocard on and set the frequency to the default station */
void s_start(int sock)
{
  char buf[BUFSIZ+1];
  Tune(defaultstation.freq,1);
  sprintf(buf,"Radiocard is on and set to the default station %s %7.3f MHz\n",defaultstation.name,defaultstation.freq);
  write(sock, buf, strlen(buf));
  return;
}

/* switch radiocard off */
void s_off(int sock)
{
  char buf[BUFSIZ+1];
  CardOff(); 
  sprintf(buf,"Radiocard is off\n");
  write(sock, buf, strlen(buf));
  return;
}

/* mute the radiocard */
void s_mute(int sock)
{
  char buf[BUFSIZ+1];
  CardMute(); 
  sprintf(buf,"Changed mute status\n");
  write(sock, buf, strlen(buf));
  return;
}

/* print list of radio stations read from config file */
void s_print(int sock)
{
  int i;
  char buf[BUFSIZ+1];
  for(i=0; i<NumStations; i++)
    {
      if(stations[i].name[0]!='\0')
	{
	  sprintf(buf,"%2d) %-10s\t- %7.3f MHz\n",i+1,stations[i].name, stations[i].freq);
	  write(sock, buf, strlen(buf));
	}
    }
  return;
}

/* volume up */
void s_up(int sock)
{
  char buf[BUFSIZ+1];
  if (lastvol<15) lastvol+=1; 
  VolumeUp(); 
  sprintf(buf,"Volume Up\n");
  write(sock, buf, strlen(buf));
  return;
}

/* volume down */
void s_down(int sock)
{
  char buf[BUFSIZ+1];
  if (lastvol>0) lastvol-=1; 
  VolumeDown(); 
  sprintf(buf,"Volume Down\n");
  write(sock, buf, strlen(buf));
  return;
}

/* kill the daemon - hidden command - if you want you can comment it out */
void s_kill(int sock)
{
  char buf[BUFSIZ+1];
  syslog(LOG_DAEMON+LOG_NOTICE,"got kill command, shutting down...");
  sprintf(buf,"Got kill command\n");
  write(sock, buf, strlen(buf));
  s_off(sock);
  close(sock);
  close(s4sig);
  closelog();
  exit(0);
}

/* tune the radiocard to one of the stations from config file */
void s_tune(int sock, char text[BUFSIZ+1])
{
  int n;
  char buf[BUFSIZ+1];
  if (text[1]==' ') 
    {
      strcpy(buf,&text[2]);
      sscanf(buf,"%d",&n);
      Tune(stations[n-1].freq,1);
      sprintf(buf,"Radio frequency set to station %s %7.3f MHz\n",stations[n-1].name, stations[n-1].freq);
      write(sock, buf, strlen(buf));
    }
  else
    if (text[4]==' ') 
      {
	strcpy(buf,&text[5]);
	sscanf(buf,"%d",&n);
	Tune(stations[n-1].freq,1);
	sprintf(buf,"Radio frequency set to station %s %7.3f MHz\n",stations[n-1].name, stations[n-1].freq);
	write(sock, buf, strlen(buf));
      }
  else
    {
      sprintf(buf,"TUNE command bad usage. Type HELP for help.\n");
      write(sock, buf, strlen(buf));
    }
  return;
}

/* tune the radiocard to the specified frequency */
void s_freq(int sock, char text[BUFSIZ+1])
{
  float freq;
  char buf[BUFSIZ+1];
  if (text[1]==' ') 
    {
      strcpy(buf,&text[2]);
      sscanf(buf,"%f",&freq);
      Tune(freq,1);
      sprintf(buf,"Radio frequency set to %7.3f\n",freq);
      write(sock, buf, strlen(buf));
    }
  else
    if (text[4]==' ') 
      {
	strcpy(buf,&text[5]);
	sscanf(buf,"%f",&freq);
	Tune(freq,1);
	sprintf(buf,"Radio frequency set to %7.3f\n",freq);
	write(sock, buf, strlen(buf));
      }
  else
    {
      sprintf(buf,"FREQ command bad usage. Type HELP for help.\n");
      write(sock, buf, strlen(buf));
    }
  return;
}

/* get the command and execute it */

void radio(int sock)
{
  char buf[BUFSIZ+1];
  int i;

  /* Say hello to the client */
  sprintf(buf, "Radio server program $Id: radiod.c,v 2.0.1.3 1997/08/05 21:36:14 anad Exp $\nWrite command or HELP for help\n\n");
  write(sock, buf, strlen(buf));
  while (1)
    {
      /* Get command */
      if ((i=read(sock, buf, BUFSIZ))<=0)
	return;
      buf[i]='\0';
      switch(tolower(buf[0])) {
        case 'q' : return; break;
	case 'h' : s_help(sock); break;
	case 's' : s_start(sock) ; break;
	case 'o' : s_off(sock); break;
	case 'm' : s_mute(sock); break;
	case 'p' : s_print(sock); break;
	case 'u' : s_up(sock); break;
	case 'd' : s_down(sock); break;
	case 't' : s_tune(sock,buf); break;
	case 'f' : s_freq(sock,buf); break;
	case '~' : s_kill(sock); break;       /* comment out 
						 if you don't want kill command */
	default : sprintf(buf,"Bad command. Type HELP for help.\n"); 
	  write(sock, buf, strlen(buf));
      }
    }
  return;
}

/* end of radiod.c */



